function SquareUnitFrames_InitialConfig(frame)
  -- Nudge the status bar frame levels down
  frame.unit.healthBar:SetFrameLevel(frame.unit:GetFrameLevel())
  frame.unit.powerBar:SetFrameLevel(frame.unit:GetFrameLevel())
  frame.unit:RegisterForDrag("LeftButton")

  frame:RegisterEvent("UNIT_HEALTH")
  frame:RegisterEvent("UNIT_MAXHEALTH")
  frame:RegisterEvent("PLAYER_ENTERING_WORLD")
  frame:RegisterEvent("UNIT_DISPLAYPOWER")
  frame:RegisterEvent("UNIT_NAME_UPDATE")
  frame:RegisterEvent("PLAYER_TARGET_CHANGED")
  frame:RegisterEvent("UNIT_THREAT_SITUATION_UPDATE")

  RegisterUnitWatch(frame.pet)
  frame.pet:SetFrameRef("unit", frame.unit)
  frame.pet:SetAttribute("_onshow", [[
  local unit = self:GetFrameRef("unit")
  unit:SetHeight(28)
  ]])
  frame.pet:SetAttribute("_onhide", [[
  local unit = self:GetFrameRef("unit")
  unit:SetHeight(34)
  ]])

  frame.pet.unit = frame.unit

  frame.pet:HookScript("OnShow", SquareUnitFrames_Pet_OnShow)
  frame.pet:HookScript("OnHide", SquareUnitFrames_Pet_OnHide)
end

function SquareUnitFrames_Frame_OnShow(button)
  local unit = button:GetAttribute("unit")

  if unit then
    local guid = UnitGUID(unit)
    if guid ~= button.guid then
      SquareUnitFrames_ResetUnitButton(button.unit, unit)
      SquareUnitFrames_ResetPetButton(button.pet, unit .. "pet")
      button.guid = guid
    end
  end
end

function SquareUnitFrames_ResetUnitButton(button, unit)
  SquareUnitFrames_ResetHealthBar(button, unit)
  SquareUnitFrames_ResetPowerBar(button, unit)
  SquareUnitFrames_ResetName(button, unit)

  if UnitIsUnit(unit, "target") then
    button.selected:Show()
  else
    button.selected:Hide()
  end

  local status = UnitThreatSituation(unit)
  if status and status > 0 then
    local r, g, b = GetThreatStatusColor(status)
    button.name:SetTextColor(r, g, b)
  else
    button.name:SetTextColor(1, 1, 1)
  end
end

-- This function can return a substring of a UTF-8 string, properly handling
-- UTF-8 codepoints.  Rather than taking a start index and optionally an end
-- index, it takes the string, the start index and the number of characters
-- to select from the string.
--
-- UTF-8 Reference:
-- 0xxxxxx - ASCII character
-- 110yyyxx - 2 byte UTF codepoint
-- 1110yyyy - 3 byte UTF codepoint
-- 11110zzz - 4 byte UTF codepoint

local function utf8sub(str, start, numChars)
  local currentIndex = start
  while numChars > 0 and currentIndex <= #str do
    local char = string.byte(str, currentIndex)
    if char >= 240 then
      currentIndex = currentIndex + 4
    elseif char >= 225 then
      currentIndex = currentIndex + 3
    elseif char >= 192 then
      currentIndex = currentIndex + 2
    else
      currentIndex = currentIndex + 1
    end
    numChars = numChars - 1
  end
  return str:sub(start, currentIndex - 1)
end

function SquareUnitFrames_ResetName(button, unit)
  local name = UnitName(unit) or "Unknown"

  local substring
  for length=#name, 1, -1 do
    substring = utf8sub(name, 1, length)
    button.name:SetText(substring)
    if button.name:GetStringWidth() <= 30 then
      return
    end
  end

  local status = UnitThreatSituation(unit)
  if status and status > 0 then
    local r, g, b = GetThreatStatusColor(status)
    button.name:SetTextColor(r, g, b)
  else
    button.name:SetTextColor(1, 1, 1)
  end
end

function SquareUnitFrames_ResetHealthBar(button, unit)
  local class = select(2, UnitClass(unit)) or "WARRIOR"
  local classColor = RAID_CLASS_COLORS[class]
  local alive = not UnitIsDeadOrGhost(unit)

  if alive then
    button.healthBar:SetStatusBarColor(classColor.r, classColor.g, classColor.b)
    button.healthBar:SetMinMaxValues(0, UnitHealthMax(unit))
    button.healthBar:SetValue(UnitHealth(unit))
    button.dead = false
  else
    button.healthBar:SetStatusBarColor(0.3, 0.3, 0.3)
    button.healthBar:SetMinMaxValues(0, 1)
    button.healthBar:SetValue(1)
    button.dead = true
  end
end

function SquareUnitFrames_ResetPowerBar(button, unit)
  local powerType, powerToken = UnitPowerType(unit)
  local powerColor = PowerBarColor[powerToken]

  local alive = not UnitIsDeadOrGhost(unit)
  if alive then
    button.powerBar:SetStatusBarColor(powerColor.r, powerColor.g, powerColor.b)
  else
    button.powerBar:SetStatusBarColor(0.3, 0.3, 0.3)
  end

  button.powerBar:SetMinMaxValues(0, UnitPowerMax(unit))
  button.powerBar:SetValue(UnitPower(unit))
end

function SquareUnitFrames_Button_OnDragStart(self, button)
  SquareUnitFrames_Header:StartMoving()
  SquareUnitFrames_Header.isMoving = true
end

function SquareUnitFrames_Button_OnDragStop(self, button)
  if SquareUnitFrames_Header.isMoving then
    SquareUnitFrames_Header:StopMovingOrSizing()
  end
end

function SquareUnitFrames_Frame_OnEvent(self, event, arg1, ...)
  local unit = self:GetAttribute("unit")
  if not unit then
    return
  end

  -- Handle any events that don't accept a unit argument
  if event == "PLAYER_ENTERING_WORLD" then
    SquareUnitFrames_ResetUnitButton(self.unit, unit)
  elseif event == "PLAYER_TARGET_CHANGED" then
    if UnitIsUnit(unit, "target") then
      self.unit.selected:Show()
    else
      self.unit.selected:Hide()
    end
  elseif arg1 and UnitIsUnit(unit, arg1) then
    local powerType, powerToken = UnitPowerType(unit)
    if event == "UNIT_MAXHEALTH" then
      if self.unit.dead ~= UnitIsDeadOrGhost(unit) then
        SquareUnitFrames_ResetUnitButton(self.unit, unit)
      else
        self.unit.healthBar:SetMinMaxValues(0, UnitHealthMax(unit))
        self.unit.healthBar:SetValue(UnitHealth(unit))
      end
    elseif event == "UNIT_HEALTH" then
      if self.unit.dead ~= UnitIsDeadOrGhost(unit) then
        SquareUnitFrames_ResetUnitButton(self.unit, unit)
      else
        self.unit.healthBar:SetValue(UnitHealth(unit))
      end
    elseif event == "UNIT_DISPLAYPOWER" then
      SquareUnitFrames_ResetPowerBar(self.unit, unit)
    elseif event == "UNIT_" .. powerToken then
      if self.unit.dead ~= UnitIsDeadOrGhost(unit) then
        SquareUnitFrames_ResetPowerBar(self.unit, unit)
      else
        self.unit.powerBar:SetValue(UnitPower(unit))
      end
    elseif event == "UNIT_MAX" .. powerToken then
      if self.unit.dead ~= UnitIsDeadOrGhost(unit) then
        SquareUnitFrames_ResetPowerBar(self.unit, unit)
      else
        self.unit.powerBar:SetMinMaxValues(0, UnitPowerMax(unit))
        self.unit.powerBar:SetValue(UnitPower(unit))
      end
    elseif event == "UNIT_NAME_UPDATE" then
      SquareUnitFrames_ResetName(self.unit, unit)
    elseif event == "UNIT_THREAT_SITUATION_UPDATE" then
      local status = UnitThreatSituation(unit)
      if status and status > 0 then
        local r, g, b = GetThreatStatusColor(status)
        self.unit.name:SetTextColor(r, g, b)
      else
        self.unit.name:SetTextColor(1, 1, 1)
      end
    end
  elseif arg1 and UnitIsUnit(unit .. "pet", arg1) then
    if event == "UNIT_MAXHEALTH" then
      self.pet.healthBar:SetMinMaxValues(0, UnitHealthMax(arg1))
      self.pet.healthBar:SetValue(UnitHealth(arg1))
    elseif event == "UNIT_HEALTH" then
      self.pet.healthBar:SetValue(UnitHealth(arg1))
    end
  end
end

local function unregisterManyEvents(frame, ...)
  for i=1, select("#", ...) do
    local event = select(i, ...)
    frame:UnregisterEvent(event)
  end
end

function SquareUnitFrames_ResetPowerBar(button, unit)
  local powerType, powerToken = UnitPowerType(unit)
  local powerColor = PowerBarColor[powerToken]
  local alive = not UnitIsDeadOrGhost(unit)

  local parent = button:GetParent()
  unregisterManyEvents(parent, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS","UNIT_ENERGY", "UNIT_RUNIC_POWER")
  unregisterManyEvents(parent, "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY", "UNIT_MAXRUNIC_POWER")

  parent:RegisterEvent("UNIT_" .. powerToken)
  parent:RegisterEvent("UNIT_MAX" .. powerToken)

  button.powerBar:SetStatusBarColor(powerColor.r, powerColor.g, powerColor.b)
  button.powerBar:SetMinMaxValues(0, UnitPowerMax(unit))
  button.powerBar:SetValue(UnitPower(unit))
end

function SquareUnitFrames_Pet_OnShow(self)
  self.unit.healthBar:SetHeight(28)
  self.unit.powerBar:Hide()
end

function SquareUnitFrames_Pet_OnHide(self)
  self.unit.healthBar:SetHeight(30)
  self.unit.powerBar:Show()
end

function SquareUnitFrames_ResetPetButton(button, unit)
  if UnitExists(unit) then
    local health = UnitHealth(unit)
    local maxHealth = UnitHealth(unit)
    button.healthBar:SetMinMaxValues(0, maxHealth)
    button.healthBar:SetValue(health)
  end
end

-- Actually show the header frame
SquareUnitFrames_Header.initialConfigFunction = SquareUnitFrames_InitialConfig
SquareUnitFrames_Header:Show()
